import pytest
from adduct_calculator.adduct_rules import AdductTransformer
from adduct_calculator.adduct_rules import PROTON_MASS

def test_mass2ion_H():
    at = AdductTransformer()
    mz = 100.
    adduct = '[M+H]+'
    ion = at.mass2ion(mz,adduct)
    assert ion == pytest.approx(mz + PROTON_MASS,0.0001)

def test_mass2ion_K():
    at = AdductTransformer()
    mz = 100.
    adduct = '[M+K]+'
    ion = at.mass2ion(mz,adduct)
    assert ion == pytest.approx(mz + 38.9631581000907,0.0001)

def test_mass2ion_M_2H():
    at = AdductTransformer()
    mz = 100.
    adduct = '[M+2H]2+'
    ion = at.mass2ion(mz,adduct)
    assert ion == pytest.approx((mz + 2*PROTON_MASS)/2,0.0001)


# mzmine-related code
import os,glob
import xml.etree.ElementTree
import numpy as np

def pick_peaks(file_list,
                xml_template = 'batch_files/PretermPilot2Reduced.xml',
                output_dir = '/Users/simon/git/pymzmine/output',
                mzmine_command = '/Users/simon/MZmine-2.40.1/startMZmine_MacOSX.command',
                force = False):
    et = xml.etree.ElementTree.parse(xml_template)
    if type(file_list) is not list:
        file_list = [file_list]
    # Loop over files in the list (just the firts three for now)
    for filename in file_list:
        print("Creating xml batch file for {}".format(filename.split(os.sep)[-1]))

        # if any output files to be produced do not already exist
        # we run. Otherwise, we only run if force = True
        # any block that produces an output file that doesn't already exist sets this to True.
        need_to_run = False

        root = et.getroot()
        for child in root:
            # Set the input filename
            if child.attrib['method'].endswith('RawDataImportModule'):
                for e in child:
                    for g in e:
                        g.text = filename # raw data file name
            # Set the mzTab export filename
            if child.attrib['method'].endswith('MzTabExportModule'):
                for e in child:
                    for g in e:
                        tag = g.tag
                        text = g.text
                        if tag == 'current_file':
                            mztab_name = os.path.join(output_dir,filename.split(os.sep)[-1].split('.')[0]+'_pp.mzTab')
                            # check if the file exists: if it already does, don't do it again (unless Force = True)
                            g.text = mztab_name
                            if not os.path.exists(mztab_name):
                                need_to_run = True
            if child.attrib['method'].endswith('GNPSExportAndSubmitModule'):
                for e in child:
                    for g in e:
                        if g.tag == 'current_file':
                            mgf_name = os.path.join(output_dir,filename.split(os.sep)[-1].split('.')[0] + '.mgf')
                            g.text = mgf_name
                            if not os.path.exists(mgf_name):
                                need_to_run = True
            if child.attrib['method'].endswith('CSVExportModule'):
                for e in child:
                    for g in e:
                        if g.tag == 'current_file':
                            csv_box_name = os.path.join(output_dir,filename.split(os.sep)[-1].split('.')[0] + '_box.csv')
                            print(csv_box_name)
                            g.text = csv_box_name
                            if not os.path.exists(csv_box_name):
                                need_to_run = True
        # write the xml file for this input file
        new_xml_name = os.path.join(output_dir,filename.split(os.sep)[-1].split('.')[0]+'.xml')
        et.write(new_xml_name)
        # Run mzmine
        
        if force or need_to_run:
            print("Running mzMine for {}".format(filename.split(os.sep)[-1]))
            os.system(mzmine_command + ' "{}"'.format(new_xml_name))
        else:
            print("Output exists for {}. Set force = True to overwrite".format(filename))



def align(mztab_folder,
            xml_template = 'batch_files/align_from_mztab.xml',
            mzmine_command = '/Users/simon/MZmine-2.40.1/startMZmine_MacOSX.command', rt_window = 0.5,
            output_csv_name = 'pp_aligned.csv',
            specific_files = None):
    # grab all the mztab files in a folder
    # mztab_folder = '/Users/simon/git/pymzmine/output'


    # if a specific set of files is provided, use those
    # otherwise use all files in the folder
    if specific_files is None:
        tab_files = glob.glob(os.path.join(mztab_folder,'*.mzTab'))
    else:
        tab_files = specific_files
    print(tab_files)


    # The output name for the aligned peaks
    csv_name = os.path.join(mztab_folder,output_csv_name)


    et = xml.etree.ElementTree.parse(xml_template)
    root = et.getroot()
    for child in root:
        # Add all the mzTab files to the import block
        if child.attrib['method'] == 'net.sf.mzmine.modules.peaklistmethods.io.mztabimport.MzTabImportModule':
            for c in child:
                if c.tag == 'parameter' and c.attrib['name'] == 'mzTab files':
                    # remove files already in xml, should just be a dummy one
                    for f in c:
                        c.remove(f)
                    for tab_file in tab_files:
                        el = xml.etree.ElementTree.Element('file')
                        el.text = tab_file
                        c.append(el)

                if c.tag == 'parameter' and c.attrib['name'] == 'Import raw data files?':
                    # Can also import the raw data with the peak lists, although it doesn't seem to be
                    # necessary for the join alignment
                    for f in c:
                        f.text = 'false'

        ###code below added for changing rt window
        if child.attrib['method'] == 'net.sf.mzmine.modules.peaklistmethods.alignment.join.JoinAlignerModule':
            # change the RT window parameter
            for c in child:

                if c.tag == 'parameter' and c.attrib['name'] == 'Retention time tolerance' and c.attrib['type'] == 'absolute' :
                    c.text = str(rt_window)
        ###

        if child.attrib['method'] == 'net.sf.mzmine.modules.peaklistmethods.io.csvexport.CSVExportModule':
            # Export the results to .csv
            for c in child:
                if c.tag == 'parameter' and c.attrib['name'] == 'Filename':
                    for f in c:
                        f.text = csv_name
    # Write the xml file
    xml_name = os.path.join(mztab_folder,'align.xml')


    et.write(xml_name)

    # Run MZ-mine
    os.system(mzmine_command + " " + xml_name)


def match_aligned_to_original(aligned_peaks,original_files,output_dir,f_idx_dict,write_file = True,original_csv_suffix = '_quant'):
    import csv
    original_csvs = [os.path.join(output_dir,original_file + original_csv_suffix + '.csv') for original_file in original_files]

    matches = {}
    for file_pos,o in enumerate(original_files):
        # Read the original .csv file
        with open(original_csvs[file_pos],'r') as f:
            reader = csv.reader(f)
            heads = next(reader)
            local_peaks = []
            for line in reader:
                id = int(line[0])
                mz = float(line[1])
                rt = float(line[2])
                intensity = float(line[3])
                local_peaks.append((id,mz,rt,intensity))
        local_peaks.sort(key = lambda x: x[3], reverse = False)
        this_idx = f_idx_dict[o]
        sub_aligned = list(filter(lambda x: x[3][this_idx]>0.0,aligned_peaks))
        sub_aligned.sort(key = lambda x: x[3][this_idx],reverse = False)

        for i,local_peak in enumerate(local_peaks):
            a_peak = sub_aligned[i]
            assert abs(local_peak[3] - a_peak[3][this_idx]) < 0.1,print(local_peak,a_peak)
            assert abs(local_peak[1] - a_peak[1]) < 0.1

            if not a_peak in matches:
                matches[a_peak] = {}
            matches[a_peak][o] = local_peak


    if write_file:
        output_file = os.path.join(output_dir,'align_links.csv')
        with open(output_file,'w') as f:
            heads = ['align ID','row m/z','row retention time'] + original_files
            writer = csv.writer(f)
            writer.writerow(heads)
            for aligned_peak in matches:
                new_row = [aligned_peak[0],aligned_peak[1],aligned_peak[2]]
                for o in original_files:
                    val = matches[aligned_peak].get(o,None)
                    if val:
                        new_row.append(val[0])
                    else:
                        new_row.append('null')
                writer.writerow(new_row)
    return matches



def align_spectra(aligned_peaks,original_files,output_dir,matches,make_mgf = False):
    from mnet_utilities import load_mgf
    from mnet import Cluster


    spec_clusters = {a:None for a in aligned_peaks}
    spectra = {}
    for o in original_files:
        mgf_file = os.path.join(output_dir,o+'.mgf')
        print("Loading peaks from {}".format(mgf_file))
        spectra[o] = load_mgf(mgf_file)
    peaks_with_spec = 0
    n_done = 0
    for ap in spec_clusters:
        for o in original_files:
            if o in matches[ap]:
                o_id = matches[ap][o][0]
                spectrum = spectra[o].get(o_id,None)
                if spectrum:
                    if spec_clusters[ap]:
                        spec_clusters[ap].add_spectrum(spectrum)
                    else:
                        spec_clusters[ap] = Cluster(spectrum,ap[0])
                        peaks_with_spec += 1
        n_done += 1
        if n_done % 1000 == 0:
            print("done {} of {}, {} peaks have spectra".format(n_done,len(aligned_peaks),peaks_with_spec))

    if make_mgf:
        mgf_name = os.path.join(output_dir,'pp_aligned.mgf')
        print("Writing {}".format(mgf_name))
        with open(mgf_name,'w') as f:
            for ap,cl in spec_clusters.items():
                if cl:
                    f.write(cl.get_mgf_string())
    return spec_clusters


def compute_cosine_aligned_peaks(aligned_peaks,original_files,output_dir,matches,spectra = None):
    # loops thorugh aligned peaks and, where two peaks that have been aligned
    # have spectra, their cosine is computed
    # for each aligned peak, the lowest cosine is kept
    from itertools import combinations
    from scoring_functions import fast_cosine

    if not matches:
        matches = match_aligned_to_original(aligned_peaks,original_files,output_dir,f_idx_dict)

    if not spectra:
        spectra = {}
        from mnet_utilities import load_mgf
        for o in original_files:
            mgf_file = os.path.join(output_dir,o+'.mgf')
            spectra[o] = load_mgf(mgf_file)
            print("Loaded {} spectra from {}".format(len(spectra[o]),o))

    worst_matches = {}
    n_done = 0
    for aligned_peak,these_matches in matches.items():
        for f1,f2 in combinations(original_files,r = 2):
            f1_id = None
            f2_id = None
            f1_spec = None
            f2_spec = None
            f1_peak = these_matches.get(f1,None)
            if f1_peak:
                f1_id = f1_peak[0]
            f2_peak = these_matches.get(f2,None)
            if f2_peak:
                f2_id = f2_peak[0]
            if f1_id:
                f1_spec = spectra[f1].get(f1_id,None)
            if f2_id:
                f2_spec = spectra[f2].get(f2_id,None)

            if f1_spec and f2_spec:
                sc,_ = fast_cosine(f1_spec,f2_spec,0.2,1)
                # store the worst pairwise score for each set
                if not aligned_peak in worst_matches:
                    worst_matches[aligned_peak] = (f1,f2,f1_id,f2_id,f1_spec,f2_spec,sc)
                else:
                    if sc < worst_matches[aligned_peak][-1]:
                        worst_matches[aligned_peak] = (f1,f2,f1_id,f2_id,f1_spec,f2_spec,sc)
        n_done += 1
        if n_done % 1000== 0:
            print(n_done,len(aligned_peaks))
    return worst_matches

def load_aligned_peaks(output_dir,aligned_csv = 'pp_aligned.csv',original_csv_suffix = '_quant'):
    import csv
    align_file = os.path.join(output_dir,aligned_csv) # file including the aligned peaks and intensities
    original_files = glob.glob(os.path.join(output_dir,'*.xml'))

    original_files = [o.split(os.sep)[-1].split('.')[0] for o in original_files]


    original_csvs = [os.path.join(output_dir,original_file + original_csv_suffix + '.csv') for original_file in original_files]
    print(original_csvs)

    aligned_peaks = []
    f_idx_dict =  {}

    original_files_filtered = []
    with open(align_file,'r') as f:
        reader = csv.reader(f)
        heads = next(reader)
        file_bits = heads[3:]
        for o in original_files:
            temp = [1 if f.startswith(o) else 0 for f in file_bits]
            try:
                f_idx_dict[o] = temp.index(1)
                original_files_filtered.append(o)
            except:
                print("{} not in list, probably ok!".format(o))
        for line in reader:
            align_id = int(line[0])
            align_mz = float(line[1])
            align_rt = float(line[2])
            intensities = tuple([float(a) for a in line[3:-1]])
            aligned_peaks.append((align_id,align_mz,align_rt,intensities))

    return aligned_peaks,original_files_filtered,f_idx_dict



def filter_duplicates(mzml_file_list,output_folder,new_output_folder,
                        min_repetitions = 10,
                        MS1_round_precision = 2,
                        score_thresh = 0.8,
                        ms2_tol = 0.1,
                        min_match_peaks = 1):
    # finds the duplicate peaks in the mzml file, loads the outputs and writes new outputs with the duplicates removed
    import csv
    for mzml_file in mzml_file_list:
        groups = find_spectral_groups(mzml_file,
                                        MS1_round_precision = MS1_round_precision,
                                        min_match_peaks = min_match_peaks,
                                        ms2_tol = ms2_tol,
                                        score_thresh = score_thresh,
                                        )
        # find and load the output files
        csv_name = os.path.join(output_folder,mzml_file.split(os.sep)[-1].split('.')[0] + '_quant.csv')
        mztab_name = os.path.join(output_folder,mzml_file.split(os.sep)[-1].split('.')[0] + '_pp.mzTab')
        new_csv_name = os.path.join(new_output_folder,mzml_file.split(os.sep)[-1].split('.')[0] + '_quant.csv')
        new_mztab_name = os.path.join(new_output_folder,mzml_file.split(os.sep)[-1].split('.')[0] + '_pp.mzTab')

        # copy the mgf files into the new folder so that things further down the pipeline will work
        mgf_name = os.path.join(output_folder,mzml_file.split(os.sep)[-1].split('.')[0] + '.mgf')
        new_mgf_name = os.path.join(new_output_folder,mzml_file.split(os.sep)[-1].split('.')[0] + '.mgf')
        os.system('cp {} {}'.format(mgf_name,new_mgf_name))
        remove_rows = set()
        with open(csv_name,'r') as f:
            reader = csv.reader(f)
            with open(new_csv_name,'w') as g:
                writer = csv.writer(g)
                # write heads
                writer.writerow(next(reader))
                row_idx = 0
                for line in reader:
                    mz = float(line[1])
                    mz = round(mz,MS1_round_precision)
                    rt = float(line[2])*60
                    remove_this = False
                    # check if this is within one of the clusters
                    if mz in groups:
                        for sub_group in groups[mz]:
                            if sub_group.n_spectra >= min_repetitions:
                                min_rt = min([s.rt for s in sub_group.spectra])
                                max_rt = max([s.rt for s in sub_group.spectra])
                                if rt >= min_rt and rt <= max_rt:
                                    remove_this = True
                                    remove_rows.add(row_idx)
                                    break
                    if not remove_this:
                        writer.writerow(line)
                    row_idx += 1
        with open(mztab_name,'r') as f:
            reader = csv.reader(f,delimiter = '\t')
            with open(new_mztab_name,'w') as g:
                writer = csv.writer(g,delimiter = '\t')
                line = next(reader)
                while len(line) == 0 or not line[0] == 'SMH':
                    writer.writerow(line)
                    line = next(reader)
                writer.writerow(line) # write the 'SMH' line
                row_idx = 0
                for line in reader:
                    if not row_idx in remove_rows:
                        writer.writerow(line)
                    row_idx += 1

def find_spectral_groups(input_file,MS1_round_precision = 2,
                            min_match_peaks = 1,
                            ms2_tol = 0.1,
                            score_thresh = 0.8,
                            verbose = False):

    from pyopenms import MSExperiment,MzMLFile
    from mnet import Spectrum,Cluster
    from scoring_functions import fast_cosine

    print("Finding groups of spectra in {}".format(input_file))
    exp = MSExperiment()
    MzMLFile().load(input_file, exp)

    clusters = {} # dictionary, key is precursor mz, value is a list of clusters

    # loop over spectra
    cl_id = 0
    n_spec = 0
    biggest_cluster_size = 0
    for ind, spectrum in enumerate(exp):
        if spectrum.getMSLevel() != 1:
            raw_precursormz = spectrum.getPrecursors()[0].getMZ()
            precursormz = round(spectrum.getPrecursors()[0].getMZ(),MS1_round_precision)
            if not precursormz in clusters:
                clusters[precursormz] = []
            # make a Spectrum object
            peaks = zip(spectrum.get_peaks()[0],spectrum.get_peaks()[1])
            s = Spectrum(peaks,input_file,ind,None,raw_precursormz,raw_precursormz,rt = spectrum.getRT())
            if len(clusters[precursormz]) == 0:
                # no clusters exist, make one
                new_cluster = Cluster(s,cl_id)
                cl_id += 1
                clusters[precursormz].append(new_cluster)
            else:
                #match_ers exist, compute the similarity between the spectrum and each
                best_score = -1
                best_pos = None
                for pos,cl in enumerate(clusters[precursormz]):
                    sc,_ = fast_cosine(s,cl,ms2_tol,min_match_peaks)
                    if sc > best_score:
                        best_score = sc
                        best_pos = pos
                if best_score >= score_thresh:
                    clusters[precursormz][pos].add_spectrum(s)
                    if clusters[precursormz][pos].n_spectra > biggest_cluster_size:
                        biggest_cluster_size = clusters[precursormz][pos].n_spectra
                else:
                    new_cluster = Cluster(s,cl_id)
                    cl_id += 1
                    clusters[precursormz].append(new_cluster)

            n_spec += 1
        if verbose and n_spec % 100 == 0:
            print("{} Masses, {} clusters, biggest cluster = {} spectra".format(len(clusters),cl_id,biggest_cluster_size))
    return clusters


def peakareas2numpy(peak_areas,tic_normalise = True):
    pakeys = list(peak_areas.keys())
    key_idx = {}
    file_idx = {}
    papos = 0
    filepos = 0
    dlist = []
    for cluster_id in peak_areas:
        these_areas = peak_areas[cluster_id]['areas']
        key_idx[cluster_id] = papos
        for f,intensity in these_areas.items():
            if not f in file_idx:
                file_idx[f] = filepos
                filepos += 1
            dlist.append((papos,file_idx[f],intensity))
        papos += 1
    from scipy.sparse import coo_matrix
    i,j,k = zip(*dlist)
    cc = coo_matrix((k,(i,j)),shape=(len(key_idx),len(file_idx)))

    pa_array = np.array(cc.todense())

    if tic_normalise:
        pa_array_tic = pa_array/pa_array.sum(axis=0)
        return pa_array,pa_array_tic,key_idx,file_idx
    else:
        return pa_array,key_idx,file_idx



# spectrum.py
# classes / methods for handling spectra

import math

class Spectrum(object):
    def __init__(self,precursor_mz,peaks):
        self.peaks = peaks
        self.peaks.sort(key = lambda x: x[0])
        self.precursor_mz = precursor_mz
        self._sqrt_normalise()
        self.n_peaks = len(self.peaks)
    
    def __str__(self):
        return str(self.precursor_mz)

    def __repr__(self):
        return self.__str__()

    def __lt__(self,other):
        return(self.precursor_mz < other.precursor_mz)


    def _sqrt_normalise(self):
        temp = []
        total = 0.0
        for mz,intensity in self.peaks:
            temp.append((mz,math.sqrt(intensity)))
            total += intensity
        norm_facc = math.sqrt(total)
        self.normalised_peaks = []
        for mz,intensity in temp:
            self.normalised_peaks.append((mz,intensity/norm_facc))


class SpectralRecord(Spectrum):
    def __init__(self,precursor_mz,peaks,metadata,original_file,spectrum_id):
        super().__init__(precursor_mz,peaks)
        self.spectrum_id = spectrum_id
        self.metadata = metadata
        self.original_file = original_file        

    def __str__(self):
        if 'names' in self.metadata:
            names = self.metadata['names'][0]
        else:
            names = ""
        return ", ".join([self.spectrum_id,names,str(self.precursor_mz),self.original_file])




import csv
class PickedBox(object):
    def __init__(self,peak_id,mz,rt,mz_min,mz_max,rt_min,rt_max,area = None,height = None):
        self.peak_id = peak_id
        self.rt = rt
        self.rt_in_seconds = rt*60.0
        self.mz = mz
        self.rt_in_minutes = rt
        self.mz_range = [mz_min,mz_max]
        self.rt_range = [rt_min,rt_max]
        self.ms2_scans = []
        self.area = area
        self.height = height
        self.rt_range_in_seconds = [60.0*r for r in self.rt_range]
    
    def __str__(self):
        return str(self.peak_id) + ": " + str(self.mz_range) + " " + str(self.rt_range)
    
    def find_ms2_scans(self,mzml_file):
        mz_range = self.mz_range
        rt_range = self.rt_range
        scans = list(filter(lambda x: 
                    x.ms_level == 2 and
                            x.precursor_mz >= mz_range[0] and
                            x.precursor_mz <= mz_range[1] and
                            x.rt_in_minutes >= rt_range[0] and
                            x.rt_in_minutes <= rt_range[1],
                    mzml_file.scans))
        self.ms2_scans = scans
        

def load_picked_boxes(csv_name):
    boxes = []
    with open(csv_name,'r') as f:
        reader = csv.reader(f)
        heads = next(reader)
        id_pos = 0
        mz_pos = 1
        rt_pos = 2
        rt_start = [i for i,h in enumerate(heads) if 'RT start' in h][0]
        rt_end = [i for i,h in enumerate(heads) if 'RT end' in h][0]
        mz_min = [i for i,h in enumerate(heads) if 'm/z min' in h][0]
        mz_max = [i for i,h in enumerate(heads) if 'm/z max' in h][0]
        try:
            area_pos = [i for i,h in enumerate(heads) if 'Peak area' in h][0]
            height_pos = [i for i,h in enumerate(heads) if 'Peak height' in h][0]
        except:
            area_pos = None
            height_pos = None
        for line in reader:
            if area_pos:
                boxes.append(PickedBox(int(line[id_pos]),
                                    float(line[mz_pos]),
                                    float(line[rt_pos]),
                                    float(line[mz_min]),
                                    float(line[mz_max]),
                                    float(line[rt_start]),
                                    float(line[rt_end]),
                                    area = float(line[area_pos]),
                                    height = float(line[height_pos])))
            else:
                boxes.append(PickedBox(int(line[id_pos]),
                                    float(line[mz_pos]),
                                    float(line[rt_pos]),
                                    float(line[mz_min]),
                                    float(line[mz_max]),
                                    float(line[rt_start]),
                                    float(line[rt_end])))

        
    return boxes


def map_boxes_to_scans(mzml_file,boxes,half_isolation_window = 0.75,allow_last_overlap = False, scan_shift_seconds = 0):
    scans2boxes = {}
    boxes2scans = {}
    for scan in mzml_file.scans:
        if scan.ms_level == 1:
            continue
        rt = scan.rt_in_minutes + scan_shift_seconds/60.
        if allow_last_overlap:
            previous_ms1 = scan.previous_ms1
            if previous_ms1:
                rt = previous_ms1.rt_in_minutes + scan_shift_seconds/60.
        min_mz = scan.precursor_mz - half_isolation_window
        max_mz = scan.precursor_mz + half_isolation_window
        sub_boxes = list(filter(lambda x: 
                        min_mz <= x.mz_range[1] and
                        max_mz >= x.mz_range[0] and
                        rt >= x.rt_range[0] and 
                        rt <= x.rt_range[1], boxes))
        if len(sub_boxes) > 0:
            scans2boxes[scan] = sub_boxes
            for box in sub_boxes:
                if not box in boxes2scans:
                    boxes2scans[box] = []
                boxes2scans[box].append(scan)
    return scans2boxes,boxes2scans

def traverse_boxes_scans(boxes2scans,scans2boxes,start_scan):
    subset_scans = set()
    subset_boxes = set()
    scans_to_explore = set()
    scans_to_explore.add(start_scan)
    while len(scans_to_explore) > 0:
        this_scan = scans_to_explore.pop()
        subset_scans.add(this_scan)
        sub_boxes = scans2boxes[this_scan]
        for b in sub_boxes:
            if not b in subset_boxes:
                subset_boxes.add(b)
                for sc in boxes2scans[b]:
                    if not sc in subset_scans:
                        scans_to_explore.add(sc)
    return subset_scans,subset_boxes


# hmdb.py
from lxml import etree
import csv
import glob
import os
from ..adduct_calculator.adduct_rules import AdductTransformer
from molmass import Formula
from .spectrum import SpectralRecord

# following method parses one of the hmdb metabolite (i.e. non msms) files
# to produce a .csv file that maps HMDB accessions to chemical formulas
# this is used when we parse msms files to assign a chemical formula, and therefore
# a precursor mz to the spectra
# Only need to run this once to produce the .csv...it takes a while
def parse_metabolites_file(xml_file_name = "/Users/simon/hmdb/hmdb_metabolites.xml",output_csv = '/Users/simon/hmdb/hmdb_metabolites.csv'):
    tree = etree.parse(xml_file_name)
    root = tree.getroot()
    acc_form = {}
    n_done = 0
    for element in root:
        n_done += 1
        try:
            accession = element.find('{http://www.hmdb.ca}accession').text
            formula = element.find('{http://www.hmdb.ca}chemical_formula').text
            acc_form[accession] = formula
        except:
            print("FAIL")
        if n_done % 100 == 0:
            print(n_done)
    with open(output_csv,'w') as f:
        writer = csv.writer(f)
        for k,v in acc_form.items():
            writer.writerow([k,v])
    return acc_form

# parses an individual msms file
# note that some have no msms peaks
# returns the spectrum ID, the HMDB ID
# the peaks, and the ionisation mode
def parse_msms_file(xml_file_name):
    tree = etree.parse(xml_file_name)
    root = tree.getroot()
    db_id = root.find('database-id').text
    mode = root.find('ionization-mode').text
    spec_id = root.find('id').text
    msms_peaks = root.find('ms-ms-peaks')
    peaks = []
    if not msms_peaks is None:
        for peak in msms_peaks:
            mz = float(peak.find('mass-charge').text)
            intensity = float(peak.find('intensity').text)
            peaks.append((mz,intensity))
    return spec_id,db_id,mode,peaks

# helper method to load the accession to formula csv
def load_csv(csv_file):
    output_dict = {}
    with open(csv_file,'r') as f:
        reader = csv.reader(f)
        for line in reader:
            output_dict[line[0]] = line[1]
    return output_dict

# load a folder of records 
# pass it the folder path
# the path to the csv mentioned above
# the target mode  ('positive','negative','both')
# and the transformation(s) you want to apply to make
# the precursor_mz values 
def load_hmdb_msms_records(folder,accession_to_formula_file,target_mode = 'positive',transformations = ['[M+H]+'],records = {}):
    accession_to_formula = load_csv(accession_to_formula_file)
    xml_files = glob.glob(os.path.join(folder,'*.xml'))
    at = AdductTransformer()
    n_loaded = 0
    n_total = len(xml_files)
    for xml_file in xml_files:
        spectrum_id,db_id,mode,peaks = parse_msms_file(xml_file)
        if len(peaks) == 0: # no peaks, ignore
            continue
        if target_mode == 'positive' and mode.lower() == 'negative': # sometimes capitalised!
            continue
        if target_mode == 'negative' and mode.lower() == 'positive':
            continue
        
        if not db_id in accession_to_formula:
            print("{} not in accession_to_formula, skipping".format(db_id))
            continue
        try:
            f = Formula(accession_to_formula[db_id])
            f_mass = f.isotope.mass
            for transformation in transformations:
                precursor_mz = at.mass2ion(f_mass,transformation)
                # make a spectral record with this as the precursor mz
                metadata = {'precursor_mz': precursor_mz,
                            'hmdb_id': db_id,
                            'mode': mode,
                            'adduct_type': transformation,
                            'spectrum_id': spectrum_id}
                ion_id = ':'.join([str(spectrum_id),db_id,transformation])
                new_record = SpectralRecord(precursor_mz,peaks,metadata,xml_file,ion_id)
                records[ion_id] = new_record
        except:
            print("Failed on fromula {}".format(accession_to_formula[db_id]))

        n_loaded += 1
        if n_loaded % 100 == 0:
            print("Loaded {} of {}".format(n_loaded,n_total))
        
    return records




import numpy as np
import pymzml
import bisect

EFFECTIVELY_INFINITE = 1e6

class MZPeak(object):
    def __init__(self,mz,intensity,scan):
        self.mz = mz
        self.intensity = intensity
        self.scan = scan


class MZMLScan(object):
    def __init__(self,scan_no,source_file,ms_level,peaks,rt_in_minutes,precursor_mz = None):
        self.scan_no = scan_no
        self.source_file = source_file
        self.ms_level = ms_level
        self.peaks = peaks
        self.rt_in_minutes = rt_in_minutes
        self.rt_in_seconds = 60.0*rt_in_minutes
        self.precursor_mz = precursor_mz
        self.previous_ms1 = None
        self.next_ms1 = None
        self.purity = None 
        self.boxes = []

        self.peaks = sorted(self.peaks,key = lambda x: x[0]) # just to be safe
    
    def compute_purity(self,ms1_scan,half_isolation_width = 0.75):
        pre = ms1_scan.get_precursor(self.precursor_mz)
        if pre:
            tot = ms1_scan.get_total_int(self.precursor_mz,half_isolation_width = half_isolation_width)
            return pre[1]/tot
        else:
            return None
    
    def get_total_int(self,precursor_mz,half_isolation_width=0.75):
        mz,intensity = zip(*self.peaks)
        mz = np.array(mz)
        intensity = np.array(intensity)
        pos1 = bisect.bisect_right(mz,precursor_mz - half_isolation_width)
        pos2 = bisect.bisect_right(mz,precursor_mz + half_isolation_width)
        return sum(intensity[pos1:pos2])

    def get_peaks_in_window(self,window_center,half_isolation_width=0.75):
        """
            A method to return all the peaks within a window

            Arguments:
            window_center -- the center of the window
            half_isolation_width -- half of the window width (i.e. all peaks
            are extracted that are with += half_isolation_width of window_center)
        """
        mz,intensity = zip(*self.peaks)
        mz = np.array(mz)
        intensity = np.array(intensity)
        pos1 = bisect.bisect_right(mz,window_center - half_isolation_width)
        pos2 = bisect.bisect_right(mz,window_center + half_isolation_width)
        peaks = []
        for pos in range(pos1,pos2):
            peaks.append((mz[pos],intensity[pos]))
        return peaks
    
    def get_precursor(self,precursor_mz,max_error = 10):
        mz,intensity = zip(*self.peaks)
        mz = np.array(mz)
        err = abs(mz - precursor_mz)
        pos = np.argmin(err)
        ppm_err = 1e6*err[pos]/precursor_mz
        if ppm_err <= max_error:
            return (mz[pos],intensity[pos])
        else:
            print("Can't find peak within {} ppm, closest = {}".format(max_error,ppm_err))
            return None

    def get_intensity(self,query_mz,dp = 2):
        # checks for exact mz match
        inte = 0.0
        b,_ = safe_round(query_mz,dp)
        for mz,i in self.peaks:
            a,_ = safe_round(mz,dp)
            if a==b:
                inte += i

        return inte

    def get_max_intensity(self, mz_min=0, mz_max=EFFECTIVELY_INFINITE):
        filtered_peaks = list(filter(lambda x: x[0] >= mz_min and x[0] <= mz_max,self.peaks))
        if len(filtered_peaks) == 0:
            return None
        else:
            mz,intensity = zip(*filtered_peaks)
            max_pos = np.argmax(np.array(intensity))
            new_peak = MZPeak(mz[max_pos],intensity[max_pos],self)
            return new_peak




def filter_spectrum(spectrum,min_intensity = 5e3):
    new_peaks = list(filter(lambda x: x[1]>=min_intensity,spectrum.peaks))
    spectrum.peaks = new_peaks


def safe_round(mz,dp):
    mz*=10**dp
    i_mz = int(mz)
    rem = mz - i_mz
    if rem >= 0.5:
        i_mz += 1
    return i_mz,i_mz/10**dp

class MZMLFile(object):
    def __init__(self,file_name):
        self.file_name = file_name
        self.scans = []
        self._load_file()
        self._link_scans()

    def _link_scans(self):
        # populates the previous ms1 and next ms1 attributes
        # find the first ms1
        for i,scan in enumerate(self.scans):
            # look back for previous ms1
            scan.previous_ms1 = self._look_back(i)
            scan.next_ms1 = self._look_forward(i)

    def _look_back(self,pos):
        # find the first ms1 scan looking back from pos
        temp_pos = pos - 1
        while temp_pos >= 0 and self.scans[temp_pos].ms_level > 1:
            temp_pos -=1
        if temp_pos >= 0:
            return self.scans[temp_pos]
        else:
            return None

    def _look_forward(self,pos):
        temp_pos = pos + 1
        while temp_pos < len(self.scans) and self.scans[temp_pos].ms_level > 1:
            temp_pos += 1
        if temp_pos < len(self.scans):
            return self.scans[temp_pos]
        else:
            return None
    
    def _load_file(self):
        reader = pymzml.run.Reader(self.file_name,obo_version = '4.1.12')
        scan_no = 0
        for scan in reader:
            rt = scan.scan_time_in_minutes()
            ms_level = scan.ms_level
            peaks = scan.peaks('centroided')
            if ms_level == 2:
                try:
                    precursor_mz = scan.selected_precursors[0]['mz']
                except:
                    precursot_mz = None
            else:
                precursor_mz = None
            self.scans.append(MZMLScan(scan_no,self.file_name,
                              ms_level,peaks,rt,precursor_mz))
            scan_no += 1
        print("Loaded {} scans".format(scan_no))
    
    def get_max_intensity(self,rt_min_seconds=0,
                               rt_max_seconds=EFFECTIVELY_INFINITE,
                               mz_min=0,mz_max=EFFECTIVELY_INFINITE):
        filtered_scans = list(filter(lambda x: x.rt_in_seconds >= rt_min_seconds and x.rt_in_seconds <= rt_max_seconds,self.scans))
        best_peak = None
        for scan in filtered_scans:
            max_peak = scan.get_max_intensity(mz_min=mz_min, mz_max=mz_max)
            if max_peak is not None:
                if best_peak is None:
                    best_peak = max_peak
                else:
                    if max_peak.intensity > best_peak.intensity:
                        best_peak = max_peak
        return best_peak.mz,best_peak.scan.rt_in_seconds,best_peak.intensity



# massbank.py
# methods and classes w.r.t massbank
import glob
import os
import sys

from .spectrum import SpectralRecord

def parse_mb_file(file_name):
    with open(file_name,'r') as f:
        in_peaks = False
        peaks = []
        metadata = {}
        names = []
        precursor_mz = None
        precursor_type = None
        smiles = None
        iupac = None
        for line in f:
            if len(line) == 0:
                continue
            line = line.rstrip()
            if line.startswith('PK$PEAK'):
                in_peaks = True
                continue
            if line.startswith('MS$FOCUSED_ION: PRECURSOR_M/Z'):
                precursor_mz_string = line.split()[-1]
                try:
                    precursor_mz = float(precursor_mz_string)
                except:
                    if '/' in precursor_mz_string:
                        precursor_mz = float(precursor_mz_string.split('/')[0])
                    else:
                        precursor_mz = None
                continue
            if line.startswith('MS$FOCUSED_ION: PRECURSOR_TYPE'):
                precursor_type_string = line.split()[-1]
                if '/' in precursor_type_string:
                    precursor_type = precursor_type_string.split('/')[0]
                else:
                    precursor_type = precursor_type_string

                continue
            if line.startswith('CH$NAME'):
                names.append(' '.join(line.split(':')[1:]))
                continue
            if line.startswith('CH$SMILES'):
                smiles = line.split()[-1]
                continue
            if line.startswith('CH$IUPAC'):
                iupac = line.split()[-1]
                continue
            if in_peaks:
                if line.startswith('//'):
                    in_peaks = False
                else:
                    mz,inte,rel_inte = line.split()
                    peaks.append((float(mz),float(inte)))
        metadata['precursor_mz'] = precursor_mz
        metadata['precursor_type'] = precursor_type
        metadata['smiles'] = smiles
        metadata['iupac'] = iupac
        metadata['names'] = names
        spectrum_id = file_name.split(os.sep)[-1].split('.')[0]
        record = SpectralRecord(precursor_mz,peaks,metadata,file_name,spectrum_id)

        return record
        

def load_folder(folder_name,records = {},verbose = True):
    file_list = glob.glob(os.path.join(folder_name,'*.txt'))
    for file_name in file_list:
        spec_id = file_name.split(os.sep)[-1].split('.')[0]
        if verbose:
            print("Loading ",spec_id)
        assert not spec_id in records
        records[spec_id] = parse_mb_file(file_name)
        records[spec_id].metadata['spec_id'] = spec_id


def load_massbank(mb_dir = '/Users/simon/git/MassBank-data/'):
    folders = glob.glob(mb_dir + '*/')
    records = {}
    for folder in folders:
        n_records = len(records)
        print("Loading records from ",folder)
        load_folder(folder,records = records,verbose = False)
        print("\t Loaded {} new records".format(len(records)-n_records))
    return records



# gnps.py
from .spectrum import SpectralRecord
def load_gnps_files(file_list):
    if type(file_list) == str:
        file_list = [file_list]
    spectra = {}
    for file_name in file_list:
        spectra = load_mgf(file_name,id_field = 'SPECTRUMID',spectra = spectra)
    return spectra


def load_mgf(mgf_name,id_field = 'SCANS',spectra = {}):
    with open(mgf_name,'r') as f:
        current_metadata = {'filename':mgf_name}
        current_peaks = []
        got_record = False
        for line in f:
            line = line.rstrip()
            if len(line) == 0:
                continue
            if line.startswith('BEGIN IONS'):
                if len(current_metadata) > 1:
                    if len(current_peaks) > 0:
                        try:
                            current_metadata['names'] = [current_metadata['COMPOUNDNAME']]
                        except:
                            pass
                        if id_field == 'SCANS':
                            id_val = int(current_metadata[id_field])
                        else:
                            id_val = current_metadata[id_field]
                        spectrum = SpectralRecord(float(current_metadata['PEPMASS']),current_peaks,current_metadata,mgf_name,id_val)
                        spectra[id_val] = spectrum
                        if len(spectra)%100 == 0:
                            print("Loaded {} spectra".format(len(spectra)))
                current_metadata = {'filename':mgf_name}
                current_peaks = []
            elif len(line.split('=')) > 1:
                # it is a metadata line
                tokens = line.split('=')
                current_metadata[tokens[0]] = "=".join(tokens[1:])
            elif not line.startswith('END IONS'):
                # it's a peak
                tokens = line.split()
                mz = float(tokens[0])
                intensity = float(tokens[1])
                current_peaks.append((mz,intensity))
    # save the last one
    if len(current_peaks) > 0:
        try:
            current_metadata['names'] = [current_metadata['COMPOUNDNAME']]
        except:
            pass
        if id_field == 'SCANS':
            id_val = int(current_metadata[id_field])
        else:
            id_val = current_metadata[id_field]
        spectrum = SpectralRecord(float(current_metadata['PEPMASS']),current_peaks,current_metadata,mgf_name,id_val)
        spectra[id_val] = spectrum
    return spectra



# spectral scoring functions

def modified_cosine_similarity(spectrum1,spectrum2,tol,min_match):
    if spectrum1.n_peaks == 0 or spectrum2.n_peaks == 0:
        return 0.0,[]

    spec1 = spectrum1.normalised_peaks
    spec2 = spectrum2.normalised_peaks

    zero_pairs = find_pairs(spec1,spec2,tol,shift=0.0)

    shift = spectrum1.precursor_mz - spectrum2.precursor_mz

    nonzero_pairs = find_pairs(spec1,spec2,tol,shift = shift)

    matching_pairs = zero_pairs + nonzero_pairs

    matching_pairs = sorted(matching_pairs,key = lambda x: x[2], reverse = True)

    used1 = set()
    used2 = set()
    score = 0.0
    used_matches = []
    for m in matching_pairs:
        if not m[0] in used1 and not m[1] in used2:
            score += m[2]
            used1.add(m[0])
            used2.add(m[1])
            used_matches.append(m)
    if len(used_matches) < min_match:
        score = 0.0
    return score,used_matches


def find_pairs(spec1,spec2,tol,shift=0):
    matching_pairs = []
    spec2lowpos = 0
    spec2length = len(spec2)
    
    for idx,(mz,intensity) in enumerate(spec1):
        # do we need to increase the lower idx?
        while spec2lowpos < spec2length and spec2[spec2lowpos][0] + shift < mz - tol:
            spec2lowpos += 1
        if spec2lowpos == spec2length:
            break
        spec2pos = spec2lowpos
        while(spec2pos < spec2length and spec2[spec2pos][0] + shift < mz + tol):
            matching_pairs.append((idx,spec2pos,intensity*spec2[spec2pos][1]))
            spec2pos += 1
        
    return matching_pairs    

def cosine_similarity(spectrum1,spectrum2,tol,min_match):
    # spec 1 and spec 2 have to be sorted by mz
    if spectrum1.n_peaks == 0 or spectrum2.n_peaks == 0:
        return 0.0,[]
    # find all the matching pairs
    
    spec1 = spectrum1.normalised_peaks
    spec2 = spectrum2.normalised_peaks
    
    matching_pairs = find_pairs(spec1,spec2,tol,shift = 0.0)
    
        
        
    matching_pairs = sorted(matching_pairs,key = lambda x:x[2],reverse = True)
    used1 = set()
    used2 = set()
    score = 0.0
    used_matches = []
    for m in matching_pairs:
        if not m[0] in used1 and not m[1] in used2:
            score += m[2]
            used1.add(m[0])
            used2.add(m[1])
            used_matches.append(m)
    if len(used_matches) < min_match:
        score = 0.0
    return score,used_matches




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mass_spec_utils", # Replace with your own username
    version="0.0.10",
    author="Simon Rogers",
    author_email="simon.d.rogers@gmail.com",
    description="Some useful MS code",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sdrogers/mass-spec-utils",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires = [
    	'pymzml',
    	'molmass',
    	'numpy',
    	'requests',
    	'sortedcontainers',
    ],
)



# alignment.py
import csv
import os
import numpy as np

class Peak(object): # todo: add MS2 information
    def __init__(self,mz,rt,intensity,source_file,source_id):
        self.mz = mz
        self.rt = rt
        self.intensity = intensity
        self.source_file = source_file
        self.source_id = source_id
        self.ms2_spectrum = None
        
    def add_ms2_spectrum(self,ms2_spectrum):
        self.ms2_spectrum = ms2_spectrum
    
    def __str__(self):
        return "{} ({}): {},{},{}".format(self.source_file,self.source_id,self.mz,self.rt,self.ms2_spectrum)

class PeakSet(object): # todo add MS2 information
    def __init__(self,peak):
        self.peaks = [peak]
        self.n_peaks = 1
        self.mean_mz = self.peaks[0].mz
        self.mean_rt = self.peaks[0].rt
        

    def add_peak(self,peak):
        self.peaks.append(peak)
        self.n_peaks += 1
        self.mean_mz = sum([x.mz for x in self.peaks])/self.n_peaks
        self.mean_rt = sum([x.rt for x in self.peaks])/self.n_peaks
    
    def _actual_mz_tolerance(self,mz,mz_tolerance_absolute,mz_tolerance_ppm):
        ppm_tolerance = mz_tolerance_ppm*self.mean_mz/1e6
        return max(ppm_tolerance,mz_tolerance_absolute)

    def is_in_box(self,peak,mz_tolerance_absolute,mz_tolerance_ppm,rt_tolerance):
        mz_tolerance = self._actual_mz_tolerance(self.mean_mz,mz_tolerance_absolute,mz_tolerance_ppm)
        if peak.mz >= self.mean_mz - mz_tolerance and peak.mz <= self.mean_mz + mz_tolerance:
            if peak.rt >= self.mean_rt - rt_tolerance and peak.rt <= self.mean_rt + rt_tolerance:
                return True
        return False

    def compute_weight(self,peak,mz_tolerance_absolute,mz_tolerance_ppm,rt_tolerance,mz_weight,rt_weight):
        mz_tolerance = self._actual_mz_tolerance(self.mean_mz,mz_tolerance_absolute,mz_tolerance_ppm)
        mz_diff = abs(peak.mz - self.mean_mz)
        rt_diff = abs(peak.rt - self.mean_rt)
        mz_score = ((1-mz_diff)/mz_tolerance)*mz_weight
        rt_score = ((1-rt_diff)/rt_tolerance)*rt_weight
        return mz_score + rt_score

    def get_intensity(self,file_name):
        # return the intensity for the peak from file file_name
        # return 0.0 if file not present
        for peak in self.peaks:
            if peak.source_file == file_name:
                return peak.intensity
        return 0.0

    def get_ms2(self,choice = 'highest_total_intensity'):
        best_ms2 = None
        best_score = 0.0
        for p in self.peaks:
            if not p.ms2_spectrum is None:
                if choice == 'highest_total_intensity':
                    score = sum([i for mz,i in p.ms2_spectrum.peaks])
                if score >= best_score:
                    best_ms2 = p.ms2_spectrum
        return best_ms2
    
  
        
class JoinAligner(object):
    def __init__(self,mz_tolerance_absolute = 0.01,
                 mz_tolerance_ppm = 10,
                 rt_tolerance = 0.5,
                 mz_column_pos = 1,
                 rt_column_pos = 2,
                 intensity_column_pos = 3):
        self.peaksets = []
        self.mz_tolerance_absolute = mz_tolerance_absolute
        self.mz_tolerance_ppm = mz_tolerance_ppm
        self.rt_tolerance = rt_tolerance
        self.mz_weight = 75
        self.rt_weight = 25
        self.files_loaded = []
        self.mz_column_pos = mz_column_pos
        self.rt_column_pos = rt_column_pos
        self.intensity_column_pos = intensity_column_pos
    def add_file(self,input_csv,input_mgf=None):
        with open(input_csv,'r') as f:
            reader =  csv.reader(f)
            heads = next(reader)
            these_peaks = []
            try:
                short_name = input_csv.split(os.sep)[-1].split('.')[0].split('_quant')[0]
            except:
                short_name = input_csv.split(os.sep)[-1]
                
            for line in reader:
                source_id = int(line[0])
                peak_mz = float(line[self.mz_column_pos])
                peak_rt = float(line[self.rt_column_pos])
                peak_intensity = float(line[self.intensity_column_pos]) 

                these_peaks.append(Peak(peak_mz,peak_rt,peak_intensity,short_name,source_id))
        self._align(these_peaks,short_name)
    
    def _align(self,these_peaks,short_name):
        if len(self.peaksets) == 0:
            # first file
            for peak in these_peaks:
                self.peaksets.append(PeakSet(peak))
        else:
            for peakset in self.peaksets:
                candidates = list(filter(lambda x: peakset.is_in_box(x,self.mz_tolerance_absolute,self.mz_tolerance_ppm,self.rt_tolerance),these_peaks))
                if len(candidates) == 0:
                    continue
                else:
                    best_peak = None
                    best_score = 0
                    for peak in candidates:
                        score = peakset.compute_weight(peak,self.mz_tolerance_absolute,self.mz_tolerance_ppm,self.rt_tolerance,self.mz_weight,self.rt_weight)
                        if score > best_score:
                            best_score = score
                            best_peak = peak
                    peakset.add_peak(best_peak)
                    pos = these_peaks.index(best_peak)
                    del these_peaks[pos]
            for peak in these_peaks: # remaining ones
                self.peaksets.append(PeakSet(peak))
        self.files_loaded.append(short_name)

    def filter_peaksets(self,min_peaks_per_peakset):
        temp = list(filter(lambda x: len(x.peaks)>= min_peaks_per_peakset,self.peaksets))
        self.peaksets = temp

    def to_matrix(self):
        n_peaksets = len(self.peaksets)
        n_files = len(self.files_loaded)
        intensity_matrix = np.zeros((n_peaksets,n_files),np.double)
        for i,peakset in enumerate(self.peaksets):
            for j,filename in enumerate(self.files_loaded):
                intensity_matrix[i,j] = peakset.get_intensity(filename)
        return intensity_matrix

class BoxJoinAligner(JoinAligner):
    def __init__(self,mz_tolerance_absolute = 0.01,
                 mz_tolerance_ppm = 10,
                 rt_tolerance = 0.5,
                 mz_column_pos = 1,
                 rt_column_pos = 2,
                 intensity_column_pos = 3):
        super().__init__(mz_tolerance_absolute,
                        mz_tolerance_ppm,
                        rt_tolerance,
                        mz_column_pos,
                        rt_column_pos,
                        intensity_column_pos)
        self.peaksets2boxes = {}

    def add_file(self,input_csv,input_mgf=None):
        from ms2_matching import load_picked_boxes
        with open(input_csv,'r') as f:
            reader =  csv.reader(f)
            heads = next(reader)
            these_peaks = []
            try:
                short_name = input_csv.split(os.sep)[-1].split('.')[0].split('_quant')[0]
            except:
                short_name = input_csv.split(os.sep)[-1]
                
            for line in reader:
                source_id = int(line[0])
                peak_mz = float(line[self.mz_column_pos])
                peak_rt = float(line[self.rt_column_pos])
                peak_intensity = float(line[self.intensity_column_pos]) 

                these_peaks.append(Peak(peak_mz,peak_rt,peak_intensity,short_name,source_id))

        # load the boxes
        temp_boxes = load_picked_boxes(input_csv)
        # convert to dictionary for speedy lookup
        temp_boxes = {box.peak_id: box for box in temp_boxes}
        n_peaksets = len(self.peaksets)
        # do the alignment
        self._align(these_peaks,short_name)

        # add boxes to the *new* peaksets
        self._add_boxes_to_peaksets(temp_boxes,short_name,n_peaksets)

    def _add_boxes_to_peaksets(self,temp_boxes,short_name,n_peaksets):
        for peakset in self.peaksets[n_peaksets:]:
            assert not peakset in self.peaksets2boxes
            peak = peakset.peaks[0] # should only be one peak in the set
            assert peak.source_file == short_name
            # get the box
            peak_id = peak.source_id
            assert peak_id in temp_boxes
            self.peaksets2boxes[peakset] = temp_boxes[peak_id]




                    


# mb_parser
import os
import glob
import math
import random
from .massbank import load_massbank
from .gnps import load_gnps_files
from .hmdb import load_hmdb_msms_records
from .spectral_scoring_functions import cosine_similarity,modified_cosine_similarity


class SpectralLibrary(object):
    def __init__(self):
        self.records = {}
        self.sorted_record_list = []

    def _candidates(self,query_mz,ms1_tol):
        from sortedcontainers import SortedList
        pmz_list = SortedList([m.precursor_mz for m in self.sorted_record_list])
        lower = query_mz - ms1_tol
        upper = query_mz + ms1_tol
        start = pmz_list.bisect(lower)
        end = pmz_list.bisect(upper)
        return self.sorted_record_list[start:end]
    
    def _dic2list(self):
        # - makes a list of the values in records
        # - removes any without a precursor mz
        # - sorts
        record_list = list(self.records.values())
        filtered_list = list(filter(lambda x: not x.precursor_mz is None,record_list))
        return sorted(filtered_list)

    def spectral_match(self,query,
            scoring_function = 'cosine',
            ms2_tol = 0.2,
            min_match_peaks = 1,
            ms1_tol = 0.2,
            score_thresh = 0.7):
        
        candidates = self._candidates(query.precursor_mz,ms1_tol)
        hits = []
        for c in candidates:
            if scoring_function == 'cosine':
                sc,_ = cosine_similarity(query,c,ms2_tol,min_match_peaks)
            elif scoring_function == 'modified cosine':
                sc,_ = modified_cosine_similarity(query,c,ms2_tol,min_match_peaks)
            else:
                print("Unrecognised scoring function: ",scoring_function)
                return None
            if sc >= score_thresh:
                hits.append((c.spectrum_id,sc,c))
        return hits


class GNPSLibrary(SpectralLibrary):
    def __init__(self,input_mgf_files):
        super().__init__()
        self.records = load_gnps_files(input_mgf_files)
        self.sorted_record_list = self._dic2list()

class MassBankLibrary(SpectralLibrary):
    def __init__(self,mb_dir = '/Users/simon/git/MassBank-data/'):
        super().__init__()
        self.records = load_massbank(mb_dir = mb_dir)
        self.sorted_record_list = self._dic2list() # sorted by precursor mz

class HMDBLibrary(SpectralLibrary):
    def __init__(self,csv_file = '/Users/simon/hmdb/hmdb_metabolites.csv',
                 msms_file_folder = '/Users/simon/hmdb/hmdb_experimental_msms_spectra',
                 mode = 'positive',
                 transformations = ['[M+H]+']):
        super().__init__()
        self.records = load_hmdb_msms_records(msms_file_folder,csv_file,target_mode = 'positive',transformations = ['[M+H]+'])
        self.sorted_record_list = self._dic2list() # sorted by precursor mz


if __name__ == '__main__':

    # # example of massbank library
    # mbl = MassBankLibrary()

    # record_list = list(mbl.records.values())
    # a = random.randrange(len(record_list))
    # example_spectrum = record_list[a]

    # hits = mbl.spectral_match(example_spectrum,score_thresh = 0.5)
    # print(example_spectrum)
    # for hit in hits:
    #     print(hit)

    # # example of HMDB
    # hmdbl = HMDBLibrary()
    
    # hits = hmdbl.spectral_match(example_spectrum,score_thresh = 0.5)
    # for hit in hits:
    #     print(hit)

    # example og GNPS
    gl = GNPSLibrary(['/Users/simon/git/molnet/lib/matched_mibig_gnps_update.mgf'])
    record_list = list(gl.records.values())
    a = random.randrange(len(record_list))
    example_spectrum = record_list[a]
    hits = gl.spectral_match(example_spectrum,score_thresh = 0.5)
    for hit in hits:
        print(hit)



import requests
import csv
from molmass import Formula
import re





PROTON_MASS = 1.00727645199076
ELECTRON_MASS = Formula('H').isotope.mass - PROTON_MASS


class Adduct(object):
    def __init__(self,name):
        self.name = name
        self.synonyms = {}
    
    def add_synonym(self,dialect,name):
        self.synonyms[dialect] = name

    def __str__(self):
        return self.name + "  (" + ", ".join(["{}:{}".format(k,v) for k,v in self.synonyms.items()]) + ")"

class AdductThesaurus(object):
    def __init__(self):
        self.adduct_dict = {}
    def add_adduct(self,adduct):
        self.adduct_dict[adduct.name] = adduct
    def get_standard_name(self,name,dialect):
        for main_name,adduct in self.adduct_dict.items():
            local_name = adduct.synonyms.get(dialect,None)
            if local_name == name:
                return main_name
        return None



class AbstractAdductTransformer(object):
    def mass2ion(self,mass,adduct_name):
        raise NotImplementedError()

    def ion2mass(self,ion_mass,adduct_name):
        raise NotImplementedError()

    def adduct_list(self,mode = 'both'):
        raise NotImplementedError()

class AdductTransformer(AbstractAdductTransformer):
    def __init__(self,local_file_folder = None):
        self.adduct_rules = self._load_adduct_files(local_file_folder = local_file_folder)

    def adduct_list(self,mode = 'both'):
        if mode == 'positive':
            return [a for a,v in self.adduct_rules.items() if v['charge'] > 0]
        elif mode == 'negative':
            return [a for a,v in self.adduct_rules.items() if v['charge'] < 0]
        else:
            return list(adduct_rules.keys())

    def mass2ion(self,mass,adduct_name):
        try:
            mass_add = self.adduct_rules[adduct_name]['mass_add']
            mass_multi = self.adduct_rules[adduct_name]['mass_multi']
        except:
            print("{} not a valid adduct, trying to parse".format(adduct_name))
            pa = ParsingAdductTransformer()
            return pa.mass2ion(mass,adduct_name)
            # return None
        return mass*mass_multi + mass_add

    def ion2mass(self,ion_mass,adduct_name):
        try:
            mass_add = self.adduct_rules[adduct_name]['mass_add']
            mass_multi = self.adduct_rules[adduct_name]['mass_multi']
        except:
            print("{} not a valid adduct, trying to parse".format(adduct_name))
            pa = ParsingAdductTransformer()
            return pa.ion2mass(ion_mass,adduct_name)
            # return None
        return (ion_mass - mass_add)/mass_multi

    def _load_adduct_files(self,local_file_folder = None,adduct_rules = {}):
        file_urls = ['https://raw.githubusercontent.com/michaelwitting/adductDefinitions/master/adducts_pos.txt','https://raw.githubusercontent.com/michaelwitting/adductDefinitions/master/adducts_neg.txt']
        if local_file_folder:
            files = glob.glob(os.path.join(local_folder,'*.txt'))
        else:
            files = None
        if not files is None:
            for file_name in files:
                with open(file_name,'r') as f:
                    reader = csv.reader(f)
                    self._parse(reader,adduct_rules)
        else:
            for file_url in file_urls:
                r = requests.get(file_url)
                decoded_content = r.content.decode('utf-8')
                reader = csv.reader(decoded_content.splitlines(),delimiter ='\t')
                self._parse(reader,adduct_rules)
        return adduct_rules


    def _parse(self,csv_reader,adduct_rules):
        heads = next(csv_reader)
        for line in csv_reader:
            adduct_transformation = line[0]
            charge = int(line[1])
            mass_add = float(line[4])
            mass_multi = float(line[5])
            adduct_rules[adduct_transformation] = {'mass_add':mass_add,'mass_multi':mass_multi,'charge':charge}
        

class ParsingAdductTransformer(AbstractAdductTransformer):
    def __init__(self):
        self.adduct_thes = AdductThesaurus()
        self._load_adducts()
    

    def mass2ion(self,mass,adduct_name,dialect = None):
        if not dialect is None:
            adduct_string = self.adduct_thes.get_standard_name(adduct_name,dialect)
            if adduct_string is None:
                return None
        else:
            adduct_string = adduct_name

        params = adduct_string_parser(adduct_string)
        if params:
            return (mass*params[0] + params[1])/abs(params[2])
        else:
            return None
    
    def ion2mass(self,mass,adduct_name,dialect = None):
        if not dialect is None:
            adduct_string = self.adduct_thes.get_standard_name(adduct_name,dialect)
            if adduct_string is None:
                return None
        else:
            adduct_string = adduct_name
        params = adduct_string_parser(adduct_string)
        if params:
            return (mass*abs(params[2]) - params[1])/params[0] # check this!
        else:
            return None
    
    def get_transform_list(self):
        return list(self.adduct_thes.keys())

    def _load_adducts(self,positive_filename = 'adduct_csv_files/Adduct definitions and synonyms - positive adducts.csv',
                           negative_filename = 'adduct_csv_files/Adduct definitions and synonyms - negative adducts.csv'):
        try:
            self._parse_csv(positive_filename)
            self._parse_csv(negative_filename)
        except:
            print("Failed to load .csv files. Assuming normal dialect.")



    def _parse_csv(self,filename):
        import csv


        with open(filename,'r') as f:
            reader = csv.reader(f)
            top_heads = next(reader)
            main_heads = next(reader)
            dialect_pos = range(5,17)
            for line in reader:
                if len(line[0]) == 0:
                    continue # no main name
                new_adduct = Adduct(line[0])
                for dpos in dialect_pos:
                    if len(line[dpos]) > 0:
                        new_adduct.add_synonym(main_heads[dpos],line[dpos])
                self.adduct_thes.add_adduct(new_adduct)
        
        

    
# obsolete methods - for deletion

def get_transform_list():
    return list(adduct_rules.keys())

def get_positive_transform_list():
    return [a for a,v in adduct_rules.items() if v['charge'] > 0]

def get_negative_transform_list():
    return [a for a,v in adduct_rules.items() if v['charge'] < 0]



def adduct_string_parser(adduct_string):
    # Step 1, access the charge
    
    # initial hacky regex check
    hit = re.search('^\[([1-9][0-9]*)?M.*\]([1-9][0-9]*)?[+-]$',adduct_string)
    if not hit:
        print("String didn't conform to expected format")
        return None

    tokens = adduct_string.split(']')
    adduct_info = tokens[0]
    charge = tokens[1]

    if charge == '+':
        charge = 1
    elif charge == '-':
        charge = -1
    elif '+' in charge:
        charge = charge.replace('+','')
        charge = int(charge)
    elif '-' in charge:
        charge = charge.replace('-','')
        charge = -int(charge)
    else:
        charge = int(charge)

    
    tokens = adduct_info.split('M')
    m_info = tokens[0]
    adduct_info = tokens[1]
    
    m_info = m_info[1:] # remove the '['
    try:
        mass_multiplier = int(m_info)
    except:
        mass_multiplier = 1
    
    plus_pos = [('+',m.start()) for m in re.finditer("\+",adduct_info)]
    minus_pos = [('-',m.start()) for m in re.finditer("\-",adduct_info)]
    
    delim_positions = sorted(plus_pos + minus_pos,key = lambda x: x[1])
    


    mass_shift = 0
    for i,(t,d) in enumerate(delim_positions):
        if i == len(delim_positions) - 1:
            end_pos = len(adduct_info)
        else:
            end_pos = delim_positions[i+1][1]
        adduct = adduct_info[d+1:end_pos]

        # check for a pre-multiplier 
        # e.g. [M+2H]2+
        # we need to add the H twice.
        # Using Formula('2H') already removes an electron
        r = re.search('^[0-9]+',adduct)
        if not r is None:
            # get the multiplier
            multiplier = int(r.group())
            # extract the remaining adduct
            adduct = adduct[r.span()[1]:]
        else:
            multiplier = 1

        # get the mass
        formula = Formula(adduct)        
        adduct_mass = formula.isotope.mass

        # subtract if necessary
        if t == '-':
            adduct_mass = -adduct_mass

        # cumulative mass shift
        mass_shift += multiplier * adduct_mass
    
    # remove / add enough electrons
    mass_shift -= charge*ELECTRON_MASS

    return (mass_multiplier,mass_shift,charge)    





if __name__ == '__main__':
    # print("E mass: ",ELECTRON_MASS)

    # a_list = get_positive_transform_list()
    # print("Positive transformations:")
    # for a in a_list:
    #     print(a)

    # a_list = get_negative_transform_list()
    # print("Negative transformations:")
    # for a in a_list:
    #     print(a)

    # for adduct in adduct_rules:
    #     print()
    #     print("Transform 100 with {}: {}".format(adduct,mass2ion(100,adduct)))
    #     print("Transform back: ",ion2mass(mass2ion(100,adduct),adduct))


    # for adduct in adduct_rules:
    #     print()
    #     print(adduct)
    #     print(adduct_string_parser(adduct))
    #     print(adduct_rules[adduct])

    # adduct_thes = parse_csv()

    # print(adduct_thes.get_standard_name('(M+ACN+H)+','Waters'))

    at = ParsingAdductTransformer()
    print(at.mass2ion(100,'[M+2H]2+'))
    print(at.mass2ion(100,'[M-2H]2-'))
    print(at.mass2ion(120,'(M+ACN+H)+',dialect='Waters'))
    print(at.mass2ion(120,'(M+ACN+H)+')) # throws error
    print(at.mass2ion(130,'[M-2H]2-'))
    print(at.ion2mass(at.mass2ion(130,'[M-2H]2-'),'[M-2H]2-'))
    
    at2 = AdductTransformer()
    print(at2.mass2ion(100,'[M+2H]2+'))
    print(at2.mass2ion(100,'[M-2H]2-'))
    print(at2.adduct_list(mode = 'negative'))
    
